#include "LynxOpenXRHMD.h"
#include "InputCoreTypes.h"
#include "LynxSpectatorScreenController.h"
#include "Modules/ModuleManager.h"
#include "OpenXRCore.h"
#include "Kismet/KismetSystemLibrary.h"

#if PLATFORM_ANDROID
#include <android_native_app_glue.h>
#include "Android/AndroidPlatformMisc.h"
#include <dlfcn.h> 
#define XR_USE_PLATFORM_ANDROID 1
#endif

#include <openxr/openxr_platform.h>

DEFINE_LOG_CATEGORY(LogLynxOpenXRHMD);

#define LOCTEXT_NAMESPACE "LynxOpenXRHMD"

const FKey SimpleController_Left_Select_Click("SimpleController_Left_Select_Click");
const FKey SimpleController_Left_Menu_Click("SimpleController_Left_Menu_Click");
const FKey SimpleController_Right_Select_Click("SimpleController_Right_Select_Click");
const FKey SimpleController_Right_Menu_Click("SimpleController_Right_Menu_Click");

FLynxOpenXRHMD::FLynxOpenXRHMD() : LoaderHandle(nullptr)
{

}

void FLynxOpenXRHMD::StartupModule()
{
	RegisterOpenXRExtensionModularFeature();

	EKeys::AddMenuCategoryDisplayInfo("SimpleController", LOCTEXT("SimpleControllerSubCategory", "Simple Controller"), TEXT("GraphEditor.PadEvent_16x"));
	EKeys::AddKey(FKeyDetails(SimpleController_Left_Select_Click, LOCTEXT("SimpleController_Left_Select_Click", "Simple Controller (L) Select"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "SimpleController"));
	EKeys::AddKey(FKeyDetails(SimpleController_Left_Menu_Click, LOCTEXT("SimpleController_Left_Menu_Click", "Simple Controller (L) Menu"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "SimpleController"));
	EKeys::AddKey(FKeyDetails(SimpleController_Right_Select_Click, LOCTEXT("SimpleController_Right_Select_Click", "Simple Controller (R) Select"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "SimpleController"));
	EKeys::AddKey(FKeyDetails(SimpleController_Right_Menu_Click, LOCTEXT("SimpleController_Right_Menu_Click", "Simple Controller (R) Menu"), FKeyDetails::GamepadKey | FKeyDetails::NotBlueprintBindableKey, "SimpleController"));
}

bool FLynxOpenXRHMD::GetCustomLoader(PFN_xrGetInstanceProcAddr* OutGetProcAddr)
{
	UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR GetCustomLoader"));

#if PLATFORM_ANDROID
	// clear errors
	dlerror();

	LoaderHandle = FPlatformProcess::GetDllHandle(TEXT("libopenxr_loader.so"));
	if (LoaderHandle == nullptr)
	{
		UE_LOG(LogLynxOpenXRHMD, Error, TEXT("Unable to load libopenxr_loader.so, error %s"), ANSI_TO_TCHAR(dlerror()));
		return false;
	}

	// clear errors
	dlerror();

	PFN_xrGetInstanceProcAddr xrGetInstanceProcAddrPtr = (PFN_xrGetInstanceProcAddr)FPlatformProcess::GetDllExport(LoaderHandle, TEXT("xrGetInstanceProcAddr"));
	if (xrGetInstanceProcAddrPtr == nullptr)
	{
		UE_LOG(LogLynxOpenXRHMD, Error, TEXT("Unable to load OpenXR xrGetInstanceProcAddr, error %s"), ANSI_TO_TCHAR(dlerror()));
		return false;
	}
	*OutGetProcAddr = xrGetInstanceProcAddrPtr;

	extern struct android_app* GNativeAndroidApp;
	PFN_xrInitializeLoaderKHR xrInitializeLoaderKHR;
	xrGetInstanceProcAddrPtr(XR_NULL_HANDLE, "xrInitializeLoaderKHR", (PFN_xrVoidFunction*)&xrInitializeLoaderKHR);
	if (xrInitializeLoaderKHR == nullptr)
	{
		UE_LOG(LogLynxOpenXRHMD, Error, TEXT("Unable to load OpenXR xrInitializeLoaderKHR"));
		return false;
	}
	XrLoaderInitInfoAndroidKHR LoaderInitializeInfoAndroid;
	LoaderInitializeInfoAndroid.type = XR_TYPE_LOADER_INIT_INFO_ANDROID_KHR;
	LoaderInitializeInfoAndroid.next = NULL;
	LoaderInitializeInfoAndroid.applicationVM = GNativeAndroidApp->activity->vm;
	LoaderInitializeInfoAndroid.applicationContext = GNativeAndroidApp->activity->clazz;
	XR_ENSURE(xrInitializeLoaderKHR((XrLoaderInitInfoBaseHeaderKHR*)&LoaderInitializeInfoAndroid));

	return true;
#else
	return false;
#endif
}

bool FLynxOpenXRHMD::IsStandaloneStereoOnlyDevice()
{
	return isStandaloneDevice;
}

bool FLynxOpenXRHMD::GetRequiredExtensions(TArray<const ANSICHAR*>& OutExtensions)
{
	return true;
}

bool FLynxOpenXRHMD::GetInteractionProfile(XrInstance InInstance, FString& OutKeyPrefix, XrPath& OutPath, bool& OutHasHaptics)
{
	UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR GetInteractionProfile"));
	OutKeyPrefix = "SimpleController";
	xrStringToPath(InInstance, "/interaction_profiles/khr/simple_controller", &OutPath);
	OutHasHaptics = true;

	return true;
}

bool FLynxOpenXRHMD::GetSpectatorScreenController(FHeadMountedDisplayBase* InHMDBase, TUniquePtr<FDefaultSpectatorScreenController>& OutSpectatorScreenController)
{
#if PLATFORM_ANDROID
	if (isStandaloneDevice)
	{
		OutSpectatorScreenController = nullptr;
		return true;
	}
	else
	{
		OutSpectatorScreenController = MakeUnique<FLynxSpectatorScreenController>(InHMDBase);
		return true;
	}
#else
	OutSpectatorScreenController = nullptr;
	return true;
#endif
}

void FLynxOpenXRHMD::AddActions(XrInstance Instance, TFunction<XrAction(XrActionType InActionType, const FName& InName, const TArray<XrPath>& InSubactionPaths)> AddAction)
{
	UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR AddActions"));

	return;
}

void FLynxOpenXRHMD::OnEvent(XrSession InSession, const XrEventDataBaseHeader* InHeader)
{
	return;
}

const void* FLynxOpenXRHMD::OnCreateInstance(class IOpenXRHMDModule* InModule, const void* InNext)
{
	UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnCreateInstance"));
	UKismetSystemLibrary::ControlScreensaver(false);
	return InNext;
}

const void* FLynxOpenXRHMD::OnGetSystem(XrInstance InInstance, const void* InNext)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnGetSystem"));
	return InNext;
}

const void* FLynxOpenXRHMD::OnCreateSession(XrInstance InInstance, XrSystemId InSystem, const void* InNext)
{
	UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnCreateSession"));
#if PLATFORM_ANDROID
	if (GRHISupportsRHIThread && GIsThreadedRendering && GUseRHIThread_InternalUseOnly)
	{
		SetRHIThreadEnabled(false, false);
	}
#endif
	return InNext;
}

const void* FLynxOpenXRHMD::OnBeginSession(XrSession InSession, const void* InNext)
{
	UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnBeginSession"));
	return InNext;
}

const void* FLynxOpenXRHMD::OnBeginFrame(XrSession InSession, XrTime DisplayTime, const void* InNext)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnBeginFrame"));
	return InNext;
}

const void* FLynxOpenXRHMD::OnBeginProjectionView(XrSession InSession, int32 InLayerIndex, int32 InViewIndex, const void* InNext)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnBeginProjectionView"));
	return InNext;
}

const void* FLynxOpenXRHMD::OnBeginDepthInfo(XrSession InSession, int32 InLayerIndex, int32 InViewIndex, const void* InNext)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnBeginDepthInfo"));
	return InNext;
}

const void* FLynxOpenXRHMD::OnEndProjectionLayer(XrSession InSession, int32 InLayerIndex, const void* InNext, XrCompositionLayerFlags& OutFlags)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnEndProjectionLayer"));
	OutFlags |= XR_COMPOSITION_LAYER_CORRECT_CHROMATIC_ABERRATION_BIT;
	//OutFlags |= XR_COMPOSITION_LAYER_BLEND_TEXTURE_SOURCE_ALPHA_BIT;
	return InNext;
}

const void* FLynxOpenXRHMD::OnEndFrame(XrSession InSession, XrTime DisplayTime, const TArray<XrSwapchainSubImage> InColorImages, const TArray<XrSwapchainSubImage> InDepthImages, const void* InNext)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnEndFrame"));
	return InNext;
}

const void* FLynxOpenXRHMD::OnSyncActions(XrSession InSession, const void* InNext)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR OnSyncActions"));
	return InNext;
}

void FLynxOpenXRHMD::PostSyncActions(XrSession InSession)
{
	//UE_LOG(LogLynxOpenXRHMD, Log, TEXT("Lynx OpenXR PostSyncActions"));
	return;
}

IMPLEMENT_MODULE(FLynxOpenXRHMD, LynxOpenXRHMD)
#undef LOCTEXT_NAMESPACE 